(*
    Ocaml XML-RPC support.
    Copyright (C) 2004 Shawn Wagner <raevnos@pennmush.org>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*)

(** Mapping XML-RPC types to ocaml types *)

(** {1 Interfacing XML-RPC and ocaml types} *)

(** {2 XML-RPC structs} *)

(** An XML-RPC struct consists of zero or more key/value pairs, where the
key is a string and the value is a [t]. You don't need to know how the
structs are implemented internally (But if you're curious, as
[(string, t) Hashtbl.t] at the moment), just use the functions
described below to get at and access pairs.
*)
type xrs

(** {2 XML-RPC types} *)

(** This type includes all the possible XML-RPC types, and the ocaml type
they corespond to. Functions listed below allow for easy extraction of
the the ocaml type, or creation of the XML-RPC type.

Note that for [`Base64], the string argument {b should not} be
actually base64-encoded by you. Use '[`RawBase64] to hold data that's
already thus encoded.
*)
type t =
    [
      `Unit
    | `Int of int32
    | `String of string
    | `Boolean of bool
    | `Double of float
    | `Base64 of string
    | `RawBase64 of string
    | `DateTime of string
    | `Array of t array
    | `Struct of xrs
    ]

(** Type of XML-RPC error information *)
type fault = {
  desc: string;
  code: int;
}
 
(** {2 Structs again} *)

(** These functions can be used to create new structs, and manipulate
existing ones -- like, say, one returned by a remote function. *)

(** Create a new struct. *)
val make_struct: unit -> xrs

(** Insert a new key/data pair into the struct *)
val add_elem: xrs -> string -> t -> unit

(** Return the data for a key. Raises [Not_found] if the element isn't
present *)
val find_elem: xrs -> string -> t


(** {2 Miscellanous routines} *)

(** This exception is thrown by many of the conversion functions in this
module when passed a different type than what they're expecting
*)
exception Bad_type

(** [escape_cdata string] turns &lt;, &gt; and &amp; into XML/HTML entities *)
val escape_cdata: string -> string

(** {2 Converting types} *)

(** {3 Ocaml to XML-RPC} *)

(** These functions take an argument of an ocaml type and return it
stored in the corresponding XML-RPC type.
*)
val xr_of_int : int -> t
val xr_of_int32 : int32 -> t
val xr_of_string : string -> t
val xr_of_bool : bool -> t
val xr_of_float : float -> t

(** [xr_of_base64] will store data that's not already base-64 encoded. *)
val xr_of_base64 : string -> t

(** But [xr_of_rawbase64] does expect its argument to already be encoded. *)
val xr_of_rawbase64 : string -> t
val xr_of_array : t array -> t
val xr_of_struct: xrs -> t

val skip: t

(** {3 XML-RPC to Ocaml} *)

(** These functions convert Ocaml reps of XML-RPC types into native Ocaml
types. If passed a {!XmlRPCTypes.t} of the wrong type for what you want
to convert it to, they will raise {!XmlRPCTypes.Bad_type} in revenge.
*)
val ml_of_int : t -> int
val ml_of_int32 : t -> int32
val ml_of_string : t -> string
val ml_of_float : t -> float
val ml_of_bool: t -> bool

(** [ml_of_base64] will return the decoded data *)
val ml_of_base64 : t -> string

(** And [ml_of_rawbase64] will return the data still base64-encoded *)
val ml_of_rawbase64 : t -> string

val ml_of_struct: t -> xrs

val ml_of_datetime: t -> string

val ml_of_fault: t -> fault

val ml_of_array : t -> t array
val ml_of_intarray: t -> int array
val ml_of_int32array: t -> int32 array
val ml_of_floatarray: t -> float array
val ml_of_datetimearray: t -> string array
val ml_of_base64array: t -> string array
val ml_of_rawbase64array: t -> string array
val ml_of_stringarray: t -> string array
val ml_of_boolarray: t -> bool array
val ml_of_structarray: t -> xrs array

(** {3 XML conversion} *)

(** These functions are used to convert to and from XML *)

(** [print_type t] returns its argument as an XML-RPC value string *)
val print_type : t -> string

(** [parse_value node] takes a PXP value node and returns its XML-RPC value. The node must be a &lt;value&gt;. *)
val parse_value: 'a Pxp_document.node -> t
